/*************************************************************************
 *
 *  ts_list.c - Library of routines for readout and buffering of 
 *                events using a JLAB Pipeline Trigger Supervisor (TS) with 
 *                a Linux VME controller in CODA 3.0
 *
 */

/* Event Buffer definitions */
#define MAX_EVENT_POOL     10
#define MAX_EVENT_LENGTH   1024*60      /* Size in Bytes */

/* Define Interrupt source and address */
#define TS_READOUT TS_READOUT_EXT_POLL  /* Poll for available data, external triggers */
#define TS_ADDR    (21<<19)          /* GEO slot 3 */

/* TDs in slot 3 and slot 19 */
#define TD_SLOT_MASK 0x80008
#define TD_SLOT_1  3
#define TD_SLOT_2  19

/* make useful TD library variables available*/
extern int nTD;


#include "dmaBankTools.h"
#include "tsprimary_list.c" /* source required for CODA */

#define BLOCKLEVEL 2
#define BUFFERLEVEL 10

extern unsigned int tsTriggerSource;
int usePulser=1;

/* function prototype */
void rocTrigger(int arg);

/****************************************
 *  DOWNLOAD
 ****************************************/
void
rocDownload()
{

  /* Setup Address and data modes for DMA transfers
   *   
   *  vmeDmaConfig(addrType, dataType, sstMode);
   *
   *  addrType = 0 (A16)    1 (A24)    2 (A32)
   *  dataType = 0 (D16)    1 (D32)    2 (BLK32) 3 (MBLK) 4 (2eVME) 5 (2eSST)
   *  sstMode  = 0 (SST160) 1 (SST267) 2 (SST320)
   */
  vmeDmaConfig(2,5,1); 

  /* Define BLock Level */
  blockLevel = BLOCKLEVEL;


  /*****************
   *   TS SETUP
   *****************/

  if(usePulser)
    tsSetTriggerSource(TS_TRIGSRC_PULSER);
  else
    tsSetTriggerSource(TS_TRIGSRC_EXT);

  /* Enable/Disable specific inputs */
  tsSetFPInput(0x1);
  tsSetGenInput(0x1);
  tsSetGTPInput(0x0);

  /* Load the default trigger table */
  tsLoadTriggerTable();
  
  /* 
   * Trigger Holdoff rules (20 ticks = 100 ns)
   */
  tsSetTriggerHoldoff(1,20,0);
  tsSetTriggerHoldoff(2,20,0);
  tsSetTriggerHoldoff(3,20,0);
  tsSetTriggerHoldoff(4,20,0);

  /* Set the sync delay width to 0x40*32 = 2.048us */
  tsSetSyncDelayWidth(0x30, 0x40, 1);

  /* 
   * Set the Block Buffer Level
   *  0:  Pipeline mode
   *  1:  One Block per readout - "ROC LOCK" mode
   *  2-255:  "Buffered" mode.
   */
  tsSetBlockBufferLevel(BUFFERLEVEL);

  /* Set a Maximum Block count before trigger autmatically disables  (0 disables block limit)*/
  tsSetBlockLimit(0);

  /* Override the busy source set in tsInit (only if TS crate running alone) */

#define USETD
#ifdef USETD

  /* Init SD Board. and add TD Slaves */
  sdInit();
  sdSetActiveVmeSlots( 1<<TD_SLOT_1 );
  sdStatus(0);

 
 /* Setup TDs - two boards in slot 3,19*/
  tdInit(0,0,0,0);

  tdGSetBlockBufferLevel(BUFFERLEVEL);

  tdAddSlave(TD_SLOT_1,4); // TI Slave - Port 4 only 

#else
  tsSetBusySource(TS_BUSY_LOOPBACK,1);
  tsAddSlave(1); // TI Slave - Connected to TI-A on TS Front Panel
#endif

/*   tsSetPrescale(0); */

  printf("rocDownload: User Download Executed\n");

}

/****************************************
 *  PRESTART
 ****************************************/
void
rocPrestart()
{
  unsigned short iflag;
  int stat;
  int islot;


  /* Set number of events per block */
  tsSetBlockLevel(blockLevel);
  printf("rocPrestart: Block Level to be broadcasted: %d\n",blockLevel);
  /* On TD's too */
  tdGSetBlockLevel(blockLevel);

  tdGStatus(0);
  tsStatus(0);
  printf("rocPrestart: User Prestart Executed\n");

}

/****************************************
 *  GO
 ****************************************/
void
rocGo()
{
  int itd, islot, tmask;

  /* Get the current Block Level */
  blockLevel = tsGetCurrentBlockLevel();
  printf("rocGo: Block Level set to %d\n",blockLevel);

  /* Enable TD module Ports that have indicated they are active */
  for (itd=0; itd<nTD; itd++) 
    {
      tdResetSlaveConfig(tdSlot(itd));
      tmask = tdGetTrigSrcEnabledFiberMask(tdSlot(itd));
      printf("TD (Slot %d) Source Enable Mask = 0x%x\n",tdSlot(itd),tmask);
      if(tmask!=0) tdAddSlaveMask(tdSlot(itd),tmask);
    }
  tdGStatus(0);

  /* Enable modules, if needed, here */
  tsStatus(0);

  /* Example: How to start internal pulser trigger */
  if(usePulser) 
    {
#ifdef RANDOMPULSER
      /* Enable Random at rate 500kHz/(2^7) = ~3.9kHz */
      tsSetRandomTrigger(1,0x7);
#else
      /* Enable fixed rate with period (ns) 120 +30*val*(1024^0) = 21.1 us (~47.4 kHz)
	 - Generated 1000 times */
      tsSoftTrig(1,1000,0x7fff,1);
#endif
    }    

}

/****************************************
 *  END
 ****************************************/
void
rocEnd()
{

  int islot;

  /* Example: How to stop internal pulser trigger */
  if(usePulser)
    {
#ifdef RANDOMPULSER
      /* Disable random trigger */
      tsDisableRandomTrigger();
#else
      /* Disable Fixed Rate trigger */
      tsSoftTrig(1,0,700,0);
#endif
    }

  tsStatus(0);

  printf("rocEnd: Ended after %d blocks\n",tsGetIntCount());
  
}

/****************************************
 *  TRIGGER
 ****************************************/
void
rocTrigger(int arg)
{
  int ii, islot;
  int stat, dCnt, len=0, idata;
  int timeout;

  /* Readout the trigger block from the TS
     Trigger Block MUST be reaodut first */
  dCnt = tsReadTriggerBlock(dma_dabufp);
  if(dCnt<=0)
    {
      logMsg("No data or error.  dCnt = %d\n",dCnt);
    }
  else
    { /* TS Data is already in a bank structure.  Bump the pointer */
      dma_dabufp += dCnt;
    }


  /* EXAMPLE: How to open a bank (type=5) and add data words by hand */
  BANKOPEN(5,BT_UI4,blockLevel);
  *dma_dabufp++ = 0xcebaf111;
  *dma_dabufp++ = tsGetIntCount();
  *dma_dabufp++ = 0xdead;
  *dma_dabufp++ = 0xcebaf222;
  BANKCLOSE;

}

void
rocCleanup()
{
  int islot=0;

}
