/*************************************************************************
 *
 *  moller_list.c  - Library of routines for the user to write for
 *                   readout and buffering of events using
 *                   a Linux VME controller.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "flexio_rol.c"
#include "jvme.h"
#include "flexioLib.h"
#include "fadcLib.h"

/* Define Interrupt source and address */
#define FLEXIO_ADDR 0x0ea0
#define FLEXIO_MODE FLEXIO_MODE_POLL
#define FLEXIO_INPUT 2 /* Number of Input Cards */
unsigned int flexio_di_cnt;
extern unsigned int flexioIntCount[2];

int vmeSyncMode = 1;   /* Allow a flag to change between Sync/Async Readout */
int dmaCheckStatus;

/* FADC STUFF */
#define FADC_ADDR 0xed0000
#define FADC_BLOCK_MODE_ASYNC         0x81      /* Asyncronous Block Mode */
#define FADC_BLOCK_MODE_SYNC          0x01      /* Syncronous Block Mode  - Wait for done*/
#define FADC_BLOCK_MODE 1
#define FADC_EVENTS_PER_BLOCK 10
#define MAX_FADC_DATA 1980
#define FADC_SUM_THRESH 9000
#define FADC_CL_PRESCALE 1
#define FADC_CR_PRESCALE 1
extern int fadcID[20];
int FA_SLOT;

/* Event Buffer definitions */
#define MAX_EVENT_POOL     400
#define MAX_EVENT_LENGTH   1024*10      /* Size in Bytes */
/* Input and Output Partitions for VME Readout */
DMA_MEM_ID vmeIN, vmeOUT;
unsigned int intCount=0;

/* function prototype */
void rocTrigger(int arg);
void rocIntAck(unsigned int card);

void
rocDownload()
{
  unsigned int res, laddr;

  flexioAckConnect(rocIntAck,0);

  /* Initialize memory partition library */
  dmaPartInit();

  /* Setup Buffer memory to store events */
  dmaPFreeAll();
  vmeIN  = dmaPCreate("vmeIN",MAX_EVENT_LENGTH,MAX_EVENT_POOL,0);
  vmeOUT = dmaPCreate("vmeOUT",0,0,0);

  dmaPStatsAll();

  /* Setup Address and data modes for DMA transfers
   *   
   *  vmeDmaConfig(addrType, dataType, sstMode);
   *
   *  addrType = 0 (A16)    1 (A24)    2 (A32)
   *  dataType = 0 (D16)    1 (D32)    2 (BLK32) 3 (MBLK) 4 (2eVME) 5 (2eSST)
   *  sstMode  = 0 (SST160) 1 (SST267) 2 (SST320)
   */

  vmeDmaConfig(2,5,1); 

  /* Reinitialize the Buffer memory */
  dmaPReInitAll();

  /* Initialize VME Interrupt interface - use defaults */
  flexioInit(FLEXIO_ADDR,FLEXIO_MODE);

  /* Initialize FADC library for Moller Operation */
  /* Requires Special "fadcMolLib.o" object loaded */
  famInit(FADC_ADDR,0xea00);
  FA_SLOT = fadcID[0];


  printf("rocDownload: User Download Executed\n");

}

void
rocPrestart()
{
  unsigned short iflag;
  int stat;

  /* Program/Init VME Modules Here */
  flexio_di_cnt = 0;

  /* Assume FADCs have been Initialized at boot or in download */
  /* Setup FADC Programming */
  faSetBlockLevel(FA_SLOT,FADC_EVENTS_PER_BLOCK);
  /*  for Block Reads */
  faEnableBusError(FA_SLOT);


  /*  Set All channel thresholds to 0 */
  faSetThreshold(FA_SLOT,0,0xffff);


  /*  Setup option 1 processing - RAW Window Data     <-- */
  /*        option 2            - RAW Pulse Data */
  /*        option 3            - Integral Pulse Data */
  /*  Setup 200 nsec latency (PL  = 50)  */
  /*  Setup  80 nsec Window  (PTW = 20) */
  /*  Setup Pulse widths of 36ns (NSB(3)+NSA(6) = 9)  */
  /*  Setup up to 1 pulse processed */
  /*  Setup for both ADC banks(0 - all channels 0-15) */
  faSetProcMode(FA_SLOT,1,50,20,3,6,1,0);

  /*  Setup Moller FPGA Processing */
  /*  See Moller Documentaion for details */
  famSetSumThresh(0,0,FADC_SUM_THRESH);
  famSetSumThresh(0,1,FADC_SUM_THRESH);

  famSetPrescale(0,0,FADC_CL_PRESCALE);
  famSetPrescale(0,1,FADC_CR_PRESCALE);


  /*  Print out status */
  faClear(FA_SLOT);
  faStatus(FA_SLOT,0);
  faMollerStatus(0);

  /* Initialize VME Interrupt variables */

  /* Connect User Trigger Routine */
  flexioIntConnect(FLEXIO_INT_VEC,FLEXIO_DEFAULT_LEVEL,rocTrigger,0);

  printf("rocPrestart: User Prestart Executed\n");

}

void
rocGo()
{
  dmaPStatsAll();
  intCount=0;
  /*  Enable Modules here */
  /*  Send Sync Reset to FADC */
  faSync(FA_SLOT);

  /* Enable Interrupts */
  flexioEnable(FLEXIO_INPUT,1); 

  /* Enable Moller */
  famEnableTest();
  famEnable();

}

void
rocEnd()
{
  int status, count;

  /* FADC Disable */
  famDisable(0);
  famDisableTest(0);

  /*  Delay before disabling Interrupts */
  taskDelay(4);

  /* Disable Interrupts */
  flexioDisable(FLEXIO_INPUT);

  flexioIntDisconnect();

  flexioStatus();

  /* FADC Event status - Is all data read out */
  faStatus(FA_SLOT,0);
  printf("\n");
  famScalStatus(1);

  printf("\nInterrupt Counts: Card 1: %8d \n",flexioIntCount[0]);
  printf("                  Card 2: %8d \n",flexioIntCount[1]);
  printf("              Double Int: %8d \n\n",flexio_di_cnt);

  printf("rocEnd: Ended after %d events\n",intCount);
  
}

void
rocTrigger(int arg)
{
  int ii, status, nscal, nwrds, dma_done = 0;
  int iwait=0;
  int nwords;
  unsigned int stat, trigval;
  unsigned int flexio_ds[2];
  int length,size;

  trigval = (unsigned int) arg;
  /* Determine which data source (or both) fired */
  if(FLEXIO_MODE==FLEXIO_MODE_INT) 
    {
      flexio_ds[0] = ((trigval&0x0000ffff) & FLEXIO_INT_PENDING);
      flexio_ds[1] = (((trigval&0xffff0000)>>16) & FLEXIO_INT_PENDING);
    }
  else 
    {
      flexio_ds[0] = ((trigval&0x0000ffff) & FLEXIO_DATA_STATUS);
      flexio_ds[1] = (((trigval&0xffff0000)>>16) & FLEXIO_DATA_STATUS);
    }
  
  /* Bump special Interrupt Counters */
  intCount++;
  if((flexio_ds[0])&&(flexio_ds[1])) 
    {
      flexio_di_cnt++;
    }
  
  /* grap an event buffer from the pool */
  GETEVENT(vmeIN,intCount);
  if(the_event== (DMANODE *)0)
    {
      printf("%s: ERROR: Out of Buffers during a trigger!\n", __FUNCTION__);
      printf("  ROC will probably crash now\n");
    }
    
  *dma_dabufp++ = LSWAP(0xda0000da);    /* Place holder for the length */
  *dma_dabufp++ = LSWAP(trigval);       /* Save FLEXIO port info for Event type*/

  /* Get scaler data first */
  if(flexio_ds[1]) 
    {
      *dma_dabufp++ = LSWAP(0xda000011);
      nscal = famReadScalers(dma_dabufp);
      if(nscal>0) 
	{
	  dma_dabufp += nscal;
	  *dma_dabufp++ = LSWAP(nscal);
	  if(nscal>16)
	    logMsg("Scalers backed up for event %d (nscal = %d)\n",intCount,nscal,0,0,0,0);

	}
      else if(nscal==0) 
	{  /* Busy - Loop until we get all data*/
	  for(ii=0;ii<10;ii++) 
	    {
	      nscal = famReadScalers(dma_dabufp);
	      if(nscal>16)
		logMsg("Scalers backed up (and delayed) for event %d (nscal = %d)\n",intCount,nscal,0,0,0,0);
	      else if(nscal>0) 
		{
		  dma_dabufp += nscal;
		  *dma_dabufp++ = LSWAP(nscal);
		  break;
		}
	      else /* nscal<0 */
		{
		  logMsg("ERROR: Reading Moller Scalers (event count = %d (%d)\n",
			 intCount,flexioIntCount[1],0,0,0,0);
		  *dma_dabufp++ = LSWAP(0xda010bad);
		  famDisable(0);
		  getchar();
		}
	      
	    }
	  if(ii>0) 
	    logMsg("Scaler ReadOut - Busy!! (ii=%d)\n",ii,0,0,0,0,0);

	}
      else /* nscal<0 */
	{
	  logMsg("ERROR: Reading Moller Scalers (event count = %d (%d)\n",
		 intCount,flexioIntCount[1],0,0,0,0);
	  *dma_dabufp++ = LSWAP(0xda010bad);
	  famDisable(0);
	  getchar();
	}

    }

  /* Check for FADC data block next */
  if(flexio_ds[0]) 
    {
      *dma_dabufp++ = LSWAP(0xda000000);
      /* Check if a Block of data is available */
      for(ii=0;ii<100;ii++) 
	{
	  stat = faBready(FA_SLOT);
	  if (stat>0) 
	    {
	      *dma_dabufp++ = LSWAP(ii);
	      break;
	    }
	}
      if(stat > 0) 
	{
	  nwrds = faReadBlock(FA_SLOT,dma_dabufp,MAX_FADC_DATA,FADC_BLOCK_MODE_SYNC);
	  if(nwrds <= 0) 
	    {
	      logMsg("ERROR: FADC (SYNC) Block Read Failed - Status 0x%x\n",nwrds,0,0,0,0,0);
	      *dma_dabufp++ = LSWAP(0xda000bad);
	      faClear(FA_SLOT);
	    } 
	  else 
	    {
	      dma_dabufp += nwrds;
	    }
	}
      else /* stat <= 0 */
	{
	  logMsg("ERROR: NO data in FADC  datascan = 0x%x\n",stat,0,0,0,0,0);
	  getchar();
	}

    }

  PUTEVENT(vmeOUT);

  /* Check if the event length is larger than expected */
  length = (((int)(dma_dabufp) - (int)(&the_event->length))>>2) - 1;
  size = the_event->part->size;
  if(length>size) 
    {
      printf("rocLib: ERROR: Event length > Buffer size (%d > %d).  Event %d\n",
	     length,size,the_event->nevent);
    }

  if(dmaPEmpty(vmeIN)) 
    {
      /* Buffers are empty.. no more triggers until one is free'd up */
      printf("WARN: vmeIN out of event buffers. (event = %d)\n",intCount);
      if(flexio_ds[0] && flexio_ds[1]) 
	{
	  flexioDoAck[0] = 0;
	  flexioDoAck[1] = 0;
	  flexioNeedAck[0] = 1;
	  flexioNeedAck[1] = 1;
	}
      else if(flexio_ds[0]) 
	{
	  flexioDoAck[0] = 0;
	  flexioNeedAck[0] = 1;
	}
      else if(flexio_ds[1]) 
	{
	  flexioDoAck[1] = 0;
	  flexioNeedAck[1] = 1;
	}
    }
      
}

/* User defined Trigger Acknowledge Routine 
   - Attach this routine to insure FlexIO is Acknowledged before the Moller FADC */
extern UINT16      flexioCsrMode[2];

void
rocIntAck(unsigned int card)
{
  if(card>2)
    {
      logMsg("%s: ERROR: Invalid argument, card = %d\n",
	     __FUNCTION__, card, 3,4,5,6);
    }

  switch(card)
    {
    case 0:
      flexioDoAck[0] = 1;
      flexioNeedAck[0] = 0;
      flexioWriteCsr(0, flexioCsrMode[0] | FLEXIO_INT_ACK);
      famMollerDataAck(0);
      break;
    case 1:
      flexioDoAck[1] = 1;
      flexioNeedAck[1] = 0;
      flexioWriteCsr(1, flexioCsrMode[1] | FLEXIO_INT_ACK);
      break;
    case 2:
    default:
      flexioDoAck[0] = 1;
      flexioDoAck[1] = 1;
      flexioNeedAck[0] = 0;
      flexioNeedAck[1] = 0;
      flexioWriteCsr(0, flexioCsrMode[0] | FLEXIO_INT_ACK);
      flexioWriteCsr(1, flexioCsrMode[1] | FLEXIO_INT_ACK);
      famMollerDataAck(0);
    }


}
